/*
//-----------------------------------------------------------------------------
OSS - OneTimePad V1.0 (C) 2003 by A. Ogut. All rights reserved. 

License to copy and use this software is granted provided that it 
is identified as the "OSS - OneTimePad" in all material mentioning 
or referencing this software or this function. 

License is also granted to make and use derivative works provided 
that such works are identified as "derived from the OSS - OneTimePad" 
in all material mentioning or referencing the derived work. 

OSS - Ogut Software Systems makes no representations concerning either 
the merchantability of this software or the suitability of this 
software for any particular purpose. It is provided "as is" 
without express or implied warranty of any kind. 

These notices must be retained in any copies of any part of this 
documentation and/or software. 
//-----------------------------------------------------------------------------
*/

/* Statistics.c : Statistikfunktionen */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>

#include "Statistics.h"
#include "Memalloc.h"


//-----------------------------------------------------------------------------------
// Berechnet statistische Daten (Mittelwerte, Abweichungen ...)
// N: Anzahl der Daten
// BufX: Datenreihe X-Achse
// BufY: Datenreihe Y-Achse (darf NULL sein -> einfache Statistik)
// pStatInfo: Zeiger auf Statistikdaten (Ergebnisrckgabe)
// return: N bei Erfolg, sonst 0
int Statistics(int N, double *BufX, double *BufY, STATINFO *pStatInfo)
{
  int       i;
  double    fN, x, xi, xmin, xmax,
            y, yi, ymin, ymax,
            xiyi, xi2, yi2, VarX, VarY,
            Kov, Korr, xm, ym, m, t, StdX, StdY, nenner;

  if (pStatInfo == NULL) return 0;
  memset(pStatInfo, 0, sizeof(STATINFO));

  if (BufX == NULL) return 0;
  if (N <= 0) return 0;
  fN = N;

  ymin = ymax = m = t = VarX = VarY = Kov = Korr =
    xi = yi = xiyi = xi2 = yi2 = StdX = StdY = 0.0;

  // Summenwerte fr X bilden
  xmin = xmax = BufX[0];
  for (i = 0; i < N; i++) {
    x   = BufX[i];
    if (x < xmin) xmin = x;
    if (x > xmax) xmax = x;
    xi  += x;
    xi2 += (x * x);
  }

  if (BufY != NULL) {
    // Summenwerte fr Y bilden
    ymin = ymax = BufY[0];
    for (i = 0; i < N; i++) {
      x   = BufX[i];
      y   = BufY[i];
      if (y < ymin) ymin = y;
      if (y > ymax) ymax = y;
      yi   += y;
      yi2  += (y * y);
      xiyi += (x * y);
    }
  }

  // Var(X), Kov(X,Y) berechnen
  if (N > 1) {
    VarX = xi2 - ((xi * xi) / fN); if (VarX < 0.0) VarX = 0.0;
    VarY = yi2 - ((yi * yi) / fN); if (VarY < 0.0) VarY = 0.0;
    Kov  = xiyi- ((xi * yi) / fN);
    VarX = VarX / (fN - 1.0);
    VarY = VarY / (fN - 1.0);
    Kov  = Kov  / (fN - 1.0);
    StdX = sqrt(VarX);
    StdY = sqrt(VarY);
  }

  // Mittelwerte berechnen;
  xm = xi / fN;
  ym = yi / fN;

  // Steigung und y-Achsenabschnitt berechnen;
  if (VarX != 0.0) {
    m = Kov / VarX;
    t = (yi - m * xi) / fN;
  }
  nenner = StdX * StdY;
  if (nenner != 0.0) Korr = Kov / nenner;

  //
  // Berechnete Daten zurckspeichern
  //
  pStatInfo->N      = N;
  pStatInfo->XMin   = xmin;
  pStatInfo->XMax   = xmax;
  pStatInfo->XSum   = xi;
  pStatInfo->XXSum  = xi2;
  pStatInfo->YMin   = ymin;
  pStatInfo->YMax   = ymax;
  pStatInfo->YSum   = yi;
  pStatInfo->YYSum  = yi2;
  pStatInfo->XYSum  = xiyi;
  pStatInfo->XM     = xm;
  pStatInfo->YM     = ym;
  pStatInfo->StdX   = StdX;
  pStatInfo->StdY   = StdY;
  pStatInfo->VarX   = VarX;
  pStatInfo->VarY   = VarY;
  pStatInfo->Kov    = Kov;
  pStatInfo->Korr   = Korr;
  pStatInfo->M      = m;
  pStatInfo->T      = t;

  return N;
}


//-----------------------------------------------------------------------------------
// Lineare Regression auf 2 int-Feldern fr x- und y-Koordinaten
void LinRegInt(int Anz, int *BufX, int *BufY,
               double *Kov, double *VarX, double *VarY, double *xm, double *ym)
{
  int i;
  long xi, yi, xiyi, xi2, yi2;

  *VarX = *VarY = *Kov = *xm = *ym = 0.0;
  *xm = BufX[0];
  if (BufY != NULL) *ym = BufY[0];
  if (Anz < 2) return;

  xi = yi = xiyi = xi2 = yi2 = 0L;

  // Summenwerte bilden
  for (i = 0; i < Anz; i++) {
    xi += (long)BufX[i];
    if (BufY != NULL) {
      yi   += (long)BufY[i];
      yi2  += (long)BufY[i] * (long)BufY[i];
      xiyi += (long)BufX[i] * (long)BufY[i];
    }
    xi2  += (long)BufX[i] * (long)BufX[i];
  }

  // Var(X), Kov(X,Y) berechnen
  *VarX = xi2  - ((xi * xi) / Anz);
  if (*VarX < 0.0) *VarX=0.0;
  *VarY = yi2  - ((yi * yi) / Anz);
  if (*VarY < 0.0) *VarY=0.0;
  *Kov  = xiyi - ((xi * yi) / Anz);
  *VarX = *VarX / (Anz - 1);
  *VarY = *VarY / (Anz - 1);
  *Kov  = *Kov  / (Anz - 1);

  // Mittelwerte berechnen;
  *xm = xi / Anz;
  *ym = yi / Anz;

  // Steigung und y-Achsenabschnitt berechnen;
//  *m = *Kov / *Var;
//  *t = (yi - *m * xi) / Anz;
} // LinRegInt


//-----------------------------------------------------------------------------------
// Startet Statistikberechnung 
// pStatInfo: Zeiger auf Statistikdaten
void StatisticsStart(STATINFO *pStatInfo)
{
  memset(pStatInfo, 0, sizeof(STATINFO));
}


//-----------------------------------------------------------------------------------
// Sammelt Daten zur Berechnung statischtischer Daten
// N: Anzahl der Daten
// BufX: Datenreihe X-Achse
// BufY: Datenreihe Y-Achse (darf NULL sein -> einfache Statistik)
// pStatInfo: Zeiger auf Statistikdaten (Ergebnisrckgabe)
// return: N bei Erfolg, sonst 0
int StatisticsUpdate(int N, double *BufX, double *BufY, STATINFO *pStatInfo)
{
  double    x, xi, xmin, xmax,
            y, yi, ymin, ymax,
            xiyi, 
            xi2, 
            yi2;

  int       i;

  //
  // Bereits berechnete Daten lesen
  //
  xmin = pStatInfo->XMin;
  xmax = pStatInfo->XMax;
  xi   = pStatInfo->XSum;
  xi2  = pStatInfo->XXSum;
  ymin = pStatInfo->YMin;
  ymax = pStatInfo->YMax;
  yi   = pStatInfo->YSum;
  yi2  = pStatInfo->YYSum;
  xiyi = pStatInfo->XYSum;

  //
  // min-max Startwerte ermitteln
  //
  if (pStatInfo->N == 0) 
  {
    xmin = xmax = BufX[0];

    if (BufY != NULL) 
    {
      ymin = ymax = BufY[0];
    }
  }


  // Summenwerte fr X bilden
  for (i = 0; i < N; i++) 
  {
    x   = BufX[i];
    if (x < xmin) xmin = x;
    if (x > xmax) xmax = x;
    xi  += x;
    xi2 += (x * x);
  }

  if (BufY != NULL) {

    // Summenwerte fr Y bilden (optional)
    for (i = 0; i < N; i++) {
      x   = BufX[i];
      y   = BufY[i];
      if (y < ymin) ymin = y;
      if (y > ymax) ymax = y;
      yi   += y;
      yi2  += (y * y);
      xiyi += (x * y);
    }
  }

  //
  // Berechnete Daten zurckspeichern
  //
  pStatInfo->N     += N;
  pStatInfo->XMin   = xmin;
  pStatInfo->XMax   = xmax;
  pStatInfo->XSum   = xi;
  pStatInfo->XXSum  = xi2;
  pStatInfo->YMin   = ymin;
  pStatInfo->YMax   = ymax;
  pStatInfo->YSum   = yi;
  pStatInfo->YYSum  = yi2;
  pStatInfo->XYSum  = xiyi;

  return N;
}


//-----------------------------------------------------------------------------------
// Berechnet statischtische Daten, die zuvor gesammelt wurden.
// pStatInfo: Zeiger auf Statistikdaten (Ergebnisrckgabe)
// return: Anzahl der Daten(paare)
int StatisticsResult(STATINFO *pStatInfo)
{
  int       N;

  double    fN, 
            xi, yi,
            xiyi, 
            xi2, yi2, 
            VarX, VarY,
            Kov, StdX, StdY, nenner;

  N = pStatInfo->N;

  if (N == 0) 
    return 0; // keine Daten ausgewertet

  //
  // Bereits berechnete Daten lesen
  //
  xi   = pStatInfo->XSum;
  xi2  = pStatInfo->XXSum;
  yi   = pStatInfo->YSum;
  yi2  = pStatInfo->YYSum;
  xiyi = pStatInfo->XYSum;

  fN = N;

  // Var(X), Kov(X,Y) berechnen
  if (N > 1) {
    VarX = xi2 - ((xi * xi) / fN); if (VarX < 0.0) VarX = 0.0;
    VarY = yi2 - ((yi * yi) / fN); if (VarY < 0.0) VarY = 0.0;
    Kov  = xiyi- ((xi * yi) / fN);
    VarX = VarX / (fN - 1.0);
    VarY = VarY / (fN - 1.0);
    Kov  = Kov  / (fN - 1.0);
    StdX = sqrt(VarX);
    StdY = sqrt(VarY);
  }

  // Mittelwerte berechnen;
  pStatInfo->XM = xi / fN;
  pStatInfo->YM = yi / fN;

  // Steigung und y-Achsenabschnitt berechnen;
  if (VarX != 0.0) {
    pStatInfo->M = Kov / VarX;
    pStatInfo->T = (yi - pStatInfo->M * xi) / fN;
  }
  nenner = StdX * StdY;
  if (nenner != 0.0) pStatInfo->Korr = Kov / nenner;

  //
  // Berechnete Daten zurckspeichern
  //
  pStatInfo->StdX   = StdX;
  pStatInfo->StdY   = StdY;
  pStatInfo->VarX   = VarX;
  pStatInfo->VarY   = VarY;
  pStatInfo->Kov    = Kov;

#if 0
  DebugOutStatistics(pStatInfo);
#endif

  return N;
}


//-----------------------------------------------------------------------------------
// Berechnet statischtische Daten(paare)
// N: Anzahl der Daten
// BufX: Datenreihe X-Achse
// BufY: Datenreihe Y-Achse (darf NULL sein -> einfache Statistik)
// pStatInfo: Zeiger auf Statistikdaten (Ergebnisrckgabe)
// return: N bei Erfolg, sonst 0
int StatisticsXY(int N, double *BufX, double *BufY, STATINFO *pStatInfo)
{
  StatisticsStart(pStatInfo);
  StatisticsUpdate(N, BufX, BufY, pStatInfo);
  return StatisticsResult(pStatInfo);
}

//-----------------------------------------------------------------------------------
// Berechnet statischtische Daten(paare)
// nBuf:      Anzahl der Daten
// dwBuf:     Datenreihe X-Achse (DWORD)
// pStatInfo: Zeiger auf Statistikdaten (Ergebnisrckgabe)
// return:    nBuf bei Erfolg, sonst 0
int StatisticsHisto(int nBuf, DWORD *dwBuf, STATINFO *pStatInfo)
{
  double * pBuf = (double *)MALLOC(sizeof(double) * nBuf);
  int      i, n;

  if (pBuf == NULL) return 0;

  for (i = 0; i < nBuf; i++)
    pBuf[i] = dwBuf[i];
  
  StatisticsStart(pStatInfo);
  StatisticsUpdate(nBuf, pBuf, NULL, pStatInfo);
  n = StatisticsResult(pStatInfo);
  FREE(pBuf);
  
  return n;
}


//-----------------------------------------------------------------------------------
// Statistik (X/Y) ber eine MessageBox anzeigen
int MessageBoxStatisticsXY(HWND hWnd, LPCTSTR lpszCaption, STATINFO *pStatInfo)
{
  char szBuf[1024];
  
  sprintf(szBuf, "N \t= %d \nXMin \t= %0.1f \nXMax \t= %0.1f \nXSum \t= %0.1f \nXXSum \t= %0.1f\nYMin \t= %0.1f\nYMax \t= %0.1f\nYSum \t= %0.1f\nYYSum \t= %0.1f\nXYSum \t= %0.1f\nXM \t= %0.1f\nYM \t= %0.1f\nStdX \t= %0.1f\nStdY \t= %0.1f\nVarX \t= %0.1f\nVarY \t= %0.1f\nKov \t= %0.1f\nKorr \t= %0.5f\nM \t= %0.5f\nT \t= %0.1f\n\n",
          pStatInfo->N,        // Anzahl der Daten
          pStatInfo->XMin,     // Minimum fr X
          pStatInfo->XMax,     // Maximum fr X
          pStatInfo->XSum,     // Summe der X
          pStatInfo->XXSum,    // Summe der X*X
          pStatInfo->YMin,     // Minimum fr Y
          pStatInfo->YMax,     // Maximum fr Y
          pStatInfo->YSum,     // Summe der Y
          pStatInfo->YYSum,    // Summe der Y*Y
          pStatInfo->XYSum,    // Summe der X*Y
          pStatInfo->XM,       // Mittelwert X (XSum/N)
          pStatInfo->YM,       // Mittelwert Y (YSum/N)
          pStatInfo->StdX,     // Standardabweichung (fr X = sqrt(VarX))
          pStatInfo->StdY,     // Standardabweichung (fr Y = sqrt(VarY))
          pStatInfo->VarX,     // Varianz X
          pStatInfo->VarY,     // Varianz Y
          pStatInfo->Kov,      // Kovarianz
          pStatInfo->Korr,     // Korrelationsfaktor
          pStatInfo->M,        // Steigung der Geraden
          pStatInfo->T         // y-Achsabschnitt
         );
  return MessageBox(hWnd, szBuf, lpszCaption, MB_OK);
}


//-----------------------------------------------------------------------------------
// Statistik (X) ber eine MessageBox anzeigen
int MessageBoxStatisticsX(HWND hWnd, LPCTSTR lpszCaption, STATINFO *pStatInfo)
{
  char szBuf[1024];
  
  sprintf(szBuf, "N \t= %d \nMin \t= %0.1f \nMax \t= %0.1f \nSum \t= %0.1f \nSum^2 \t= %0.1f\nMean \t= %0.1f\nStd \t= %0.1f\nVar \t= %0.1f\n\n",
          pStatInfo->N,        // Anzahl der Daten
          pStatInfo->XMin,     // Minimum fr X
          pStatInfo->XMax,     // Maximum fr X
          pStatInfo->XSum,     // Summe der X
          pStatInfo->XXSum,    // Summe der X*X
          pStatInfo->XM,       // Mittelwert X (XSum/N)
          pStatInfo->StdX,     // Standardabweichung (fr X = sqrt(VarX))
          pStatInfo->VarX      // Varianz X
         );
  return MessageBox(hWnd, szBuf, lpszCaption, MB_OK);
}


//-----------------------------------------------------------------------------------
#ifdef _DEBUG
void DebugOutStatistics(STATINFO *pStatInfo)
{
  char szBuf[1024];
  
  OutputDebugString("\nStatistics:\n");
  sprintf(szBuf, "N = %d \nXMin = %0.1f \nXMax = %0.1f \nXSum = %0.1f \nXXSum = %0.1f\nYMin = %0.1f\nYMax = %0.1f\nYSum = %0.1f\nYYSum = %0.1f\nXYSum = %0.1f\nXM = %0.1f\nYM = %0.1f\nStdX = %0.1f\nStdY = %0.1f\nVarX = %0.1f\nVarY = %0.1f\nKov = %0.1f\nKorr = %0.5f\nM = %0.5f\nT = %0.1f\n\n",
          pStatInfo->N,        // Anzahl der Daten
          pStatInfo->XMin,     // Minimum fr X
          pStatInfo->XMax,     // Maximum fr X
          pStatInfo->XSum,     // Summe der X
          pStatInfo->XXSum,    // Summe der X*X
          pStatInfo->YMin,     // Minimum fr Y
          pStatInfo->YMax,     // Maximum fr Y
          pStatInfo->YSum,     // Summe der Y
          pStatInfo->YYSum,    // Summe der Y*Y
          pStatInfo->XYSum,    // Summe der X*Y
          pStatInfo->XM,       // Mittelwert X (XSum/N)
          pStatInfo->YM,       // Mittelwert Y (YSum/N)
          pStatInfo->StdX,     // Standardabweichung (fr X = sqrt(VarX))
          pStatInfo->StdY,     // Standardabweichung (fr Y = sqrt(VarY))
          pStatInfo->VarX,     // Varianz X
          pStatInfo->VarY,     // Varianz Y
          pStatInfo->Kov,      // Kovarianz
          pStatInfo->Korr,     // Korrelationsfaktor
          pStatInfo->M,        // Steigung der Geraden
          pStatInfo->T         // y-Achsabschnitt
         );
  OutputDebugString(szBuf);
}

#endif

